#include <gb.h>
#include <stdio.h>

#include "EMSAsm.h"

#include "terminal.c"


#define VER_MAJ 1
#define VER_MIN 3

#define MENU_STARTY 2

unsigned char palTiles[] =
{
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,
  0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
};



//this is the palette that will get set up.
//  if we're in color mode, then the BW games will get this palette
UWORD paletteCGB[4];
UWORD hardPaletteCGB[4] =
		{ RGB(31,31,31), RGB(20,20,20), RGB(10,10,10), RGB(0,0,0)};

struct _romInfo
{
   char romName[20];
   UBYTE cartRAM;	//Cart RAM($80)/No Cart RAM($00)
   UBYTE cartROM;	//ROM Usage (0=32k,1=65k,3=128k,7=256k,f=512k,1f=1M,3f=2M)
   UBYTE RAMsize;	//RAM Usage (0=8k,3=32k)
   UBYTE RAMoffset;	//RAM bank 0 offset in SRAM (8k byte increments)
   UBYTE ROMoffset;	//ROM bank 0 offset in Flash (32k byte increments)
} romInfo[16], savedRomInfo[16];

//new stuff for Auto SRAM
struct _SRAMInfo
{
   char ramName[20];
   UBYTE RAMbanks;  //how many total SRAM banks it uses.
} SRAMInfo[14];  //there are 14 possible SRAM banks (well, 15, but we're usin' the 15th)

UBYTE ramOverflow;      //gets set if we don't have enough SRAM for all the ROMs
UBYTE lastRunROMNumber; //pointer to the ROM you ran last

UBYTE sramMode;         //AUTO or MANUAL
#define SRAM_MODE_MANUAL 0
#define SRAM_MODE_AUTO 1

char paletteString[] = "Palette\0";
char ramMenuName[]   = "ramMenu\0";
char *RAM_B000 = (char *)0xB000;
char *RAM_B010 = (char *)0xB010;
char *RAM_B400 = (char *)0xB400;
char *RAM_BF00 = (char *)0xBF00;
char *RAM_BF10 = (char *)0xBF10;
char *RAM_BFFF = (char *)0xBFFF;

//end Auto SRAM stuff


char arrowString[] = {26, 0}; //right arrow
char upArrowString[] = {24, 0}; //right arrow

void waitJoyPress()
{
   while(!joypad()) {}
   waitpadup();
}

void printString(char x, char y, char *string)
{
   UBYTE strLen;

   strLen = strlen(string);

   set_bkg_tiles(x,y,strLen,1,(unsigned char *)string);
}

void hideSprites()
{
   UBYTE i;
   for(i = 0; i < 8; i++)
   {
      move_sprite(i, 0, 0);//hide the sprites
   }
}

void loadSRAMInfo()
{
   UBYTE i;

   //if there's one saved, use it, otherwise, use hardPaletteCGB (hard coded version)
   ENABLE_RAM_MBC5;
   SWITCH_RAM_MBC5(0x0F);

   if(strncmp(RAM_BF00, paletteString, 7) == 0)
   {
      memcpy(paletteCGB, RAM_BF10, sizeof(paletteCGB));
   }
   else
   {
      memcpy(paletteCGB, hardPaletteCGB, sizeof(paletteCGB));
   }

   //get the save RAM info from SRAM bank 0x0F
   if(strncmp(ramMenuName, RAM_B000, 7) == 0)
   {
      memcpy(savedRomInfo[0].romName, RAM_B010, sizeof(romInfo));
      memcpy(SRAMInfo, RAM_B400, sizeof(SRAMInfo));

      lastRunROMNumber = RAM_BFFF[0];
   }
   else
   {
      strcpy(SRAMInfo[0].ramName, "Current SRAM\0");
      strcpy(SRAMInfo[1].ramName, "Current SRAM\0");
      strcpy(SRAMInfo[2].ramName, "Current SRAM\0");
      strcpy(SRAMInfo[3].ramName, "Current SRAM\0");

      for(i = 4; i < 14; i++)
      {
         strcpy(SRAMInfo[i].ramName, "Empty\0");
      }
      lastRunROMNumber = 0;
   }

   SWITCH_RAM_MBC5(0x00);
   DISABLE_RAM_MBC5;
}

void saveRomAndRamInfo()
{
   UBYTE i;

   ENABLE_RAM_MBC5;
    SWITCH_RAM_MBC5(0x0F);
      for(i = 0; i < 7; i++)
         RAM_B000[i] = ramMenuName[i];
      RAM_B000[8] = 0;

      memcpy(RAM_B010, romInfo[0].romName, sizeof(romInfo));
      memcpy(RAM_B400, SRAMInfo, sizeof(SRAMInfo));
    SWITCH_RAM_MBC5(0);
   DISABLE_RAM_MBC5;
}

void savePalToRAM()
{
   UBYTE i;

   ENABLE_RAM_MBC5;
    SWITCH_RAM_MBC5(0x0F);
      for(i = 0; i < 7; i++)
         RAM_BF00[i] = paletteString[i];
      RAM_BF00[8] = 0;

      memcpy(RAM_BF10, paletteCGB, sizeof(paletteCGB));
    SWITCH_RAM_MBC5(0);
   DISABLE_RAM_MBC5;
}

void saveEverythingToRAM()
{
   saveRomAndRamInfo();
   savePalToRam();
}

void saveLastRunROMNumber(UBYTE ROMNumber)
{
   ENABLE_RAM_MBC5;
    SWITCH_RAM_MBC5(0x0F);
      RAM_BFFF[0] = ROMNumber;
    SWITCH_RAM_MBC5(0);
   DISABLE_RAM_MBC5;
}

void copyRAMbanks(UBYTE toBank, UBYTE fromBank, UBYTE numBanks)
{
   UBYTE i, j; 

   ENABLE_RAM_MBC5;
   for(i = 0; i < numBanks; i++)
   {
      //copy the first 0x1000 bytes
      SWITCH_RAM_MBC5(fromBank + i);
      memcpy((UBYTE *) 0xC500, (UBYTE *)0xA000, 0x1000);
      SWITCH_RAM_MBC5(toBank + i);
      memcpy((UBYTE *)0xA000, (UBYTE *) 0xC500, 0x1000);

      //copy the second 0x1000 bytes
      SWITCH_RAM_MBC5(fromBank + i);
      memcpy((UBYTE *) 0xC500, (UBYTE *)0xB000, 0x1000);
      SWITCH_RAM_MBC5(toBank + i);
      memcpy((UBYTE *)0xB000, (UBYTE *) 0xC500, 0x1000);
   }
   DISABLE_RAM_MBC5;
}

void doSRamSwap(UBYTE ROMNumber)
{
	int i, offset;

   //We need to figure out if the ROM that ROMNumber points to uses the currently
   // loaded SRAM.  If it does, then just return.  (or back it up and return)
   //If it doesn't, we need to back it up (to it's offset) and then load the SRAM
   // that ROMNumber needs.

   if(lastRunROMNumber == 0)  //0 is the Multi-ROM, so it's obviously bogus
      return;

   offset = romInfo[lastRunROMNumber].RAMoffset;
   
   if(ROMNumber == lastRunROMNumber)
   {
      //it's the last one we used, so just back it up
      copyRAMbanks(offset, 0, romInfo[lastRunROMNumber].RAMsize+1);
   }
   else
   {
      //we need to back it up and swap in the new one

      //back up the current one
      copyRAMbanks(offset, 0, romInfo[lastRunROMNumber].RAMsize+1);

   	for(i=0; i < SRAMInfo[0].RAMbanks; i++)
	   {
   	   strncpy(SRAMInfo[offset+i].ramName, SRAMInfo[i].ramName, 19);
	      SRAMInfo[offset+i].ramName[19] = 0;
      	SRAMInfo[offset+i].RAMbanks = 0;//zero 'em out
   	}
	   SRAMInfo[offset].RAMbanks = SRAMInfo[0].RAMbanks;//how many total SRAM banks it uses


      //load the one we need
      copyRAMbanks(0, romInfo[ROMNumber].RAMoffset, romInfo[ROMNumber].RAMsize+1);
   }

   for(i=0; i < 4; i++)
   {
      strncpy(SRAMInfo[i].ramName, romInfo[ROMNumber].romName, 19);
      SRAMInfo[i].ramName[19] = 0;
      SRAMInfo[i].RAMbanks = 0;//zero 'em out
   }
   SRAMInfo[0].RAMbanks = romInfo[ROMNumber].RAMsize+1;//how many total SRAM banks it uses
}

void backupSRam()
{
   int i, offset;
   //if the last run rom used SRAM, back it up
   if(lastRunROMNumber == 0)  //0 is the Multi-ROM, so it's obviously bogus
      return;

   if(romInfo[lastRunROMNumber].cartRAM != 0)
   {
      offset = romInfo[lastRunROMNumber].RAMoffset;
      copyRAMbanks(offset, 0, romInfo[lastRunROMNumber].RAMsize+1);

   	for(i=0; i < SRAMInfo[0].RAMbanks; i++)
	   {
   	   strncpy(SRAMInfo[offset+i].ramName, SRAMInfo[i].ramName, 19);
	      SRAMInfo[offset+i].ramName[19] = 0;
      	SRAMInfo[offset+i].RAMbanks = 0;//zero 'em out
   	}
	   SRAMInfo[offset].RAMbanks = SRAMInfo[0].RAMbanks;//how many total SRAM banks it uses
   }
}


void ramMenu()
{
   char tempString[20];
   UBYTE i, RAMposition;
   int joy;

//This is just for testing.  It will be set to Auto-Only mode to begin with
hideSprites();
cls();
printString(0, 0,  "     RAM Manager    \0");
printString(0, 1,  "is disabled for this\0");
printString(0, 2,  "    test release.   \0");
printString(0, 3,  "In a future release,\0");
printString(0, 4,  "I plan to add an    \0");
printString(0, 5,  "option to change    \0");
printString(0, 6,  "between this mode   \0");
printString(0, 7,  "(which automatically\0");
printString(0, 8,  "switches your SRAM) \0");
printString(0, 9,  "and manual mode     \0");
printString(0, 10, "(which will be like \0");
printString(0, 11, "the SRAM Manager in \0");
printString(0, 12, "version 1.1).       \0");
printString(0, 13, "Please, if you test \0");
printString(0, 14, "this automatic mode,\0");
printString(0, 15, "EMail comments to me\0");

printString(0, 17, "Flavor@Flavor.8k.com\0");
waitJoyPress();
return;

   hideSprites();
   cls();

   RAMposition = 1;

   //set_bkg_palette(0, 1, hardPaletteCGB);  //make sure we can read stuff

   printString(0, 0, "     RAM Manager    \0");
   printString(0, 2, "START-exit          \0");
   printString(0, 3, "A-save to           \0");
   printString(0, 4, "B-load from         \0");
   printString(2, 5, "--------------------\0");
   printString(2, 6, "                    \0");
   printString(2, 6, SRAMInfo[0].ramName);
   printString(2, 7, "--------------------\0");
   for(i = 1; i < 4; i++)
   {
      printString(2, 7+i, "                    \0");
      printString(2, 7+i, SRAMInfo[i].ramName);
      sprintf(tempString, "%d\0", i);
      printString(1, 7+i, tempString);
   }

   printString(0, 12,"The 3rd Bank may get\0");
   printString(0, 13,"corrupted, if used  \0");
   printString(0, 14,"with big (32k) SRAMs\0");


   joy = 0;
   printString(0, 7+RAMposition, arrowString);

   while(! (joy & J_START))
   {
      joy = joypad();

      if(joy)
      {
         //take the warning away
         printString(0, 12,"                    \0");
         printString(0, 13,"                    \0");
         printString(0, 14,"                    \0");
      }

      if(joy & (J_DOWN | J_SELECT | J_UP))
      {
         printString(0, 7+RAMposition, " \0");

         if(joy & J_UP)
            RAMposition--;
         else
            RAMposition++;

         if(RAMposition > 3)
            RAMposition = 1;
         else if(RAMposition < 1)
            RAMposition = 3;
         printString(0, 7+RAMposition, arrowString);
      }
      else if(joy & J_A)
      {
         waitpadup();
         for(i = 0; i < 20; i++)
            tempString[i] = 0;
         strcpy(tempString, SRAMInfo[0].ramName);

         printString(0, 14, "Press A to save\0");
         printString(0, 15, "Use Pad to name RAM\0");
         printString(0, 16, tempString);

         joy = 0;
         i = 0;
         printString(i, 17, upArrowString);
         while(!(joy & (J_B | J_A | J_START | J_SELECT)))
         {
            joy = joypad();

            if(joy & J_UP)
            {
               if(tempString[i] >= 127)
                  tempString[i] = 1;
               else
                  tempString[i]++;
            }
            else if(joy & J_DOWN)
            {
               if(tempString[i] <= 1)
                  tempString[i] = 126;
               else
                  tempString[i]--;
            }
            else if(joy & J_RIGHT)
            {
               if(tempString[i] != 0)//0 is the last position
               {
                  printString(0, 17, "                    \0");
                  if(i <= 16)
                     i++;
                  printString(i, 17, upArrowString);
               }
            }
            else if(joy & J_LEFT)
            {
               printString(0, 17, "                    \0");
               if(i >= 1)
                  i--;
               printString(i, 17, upArrowString);
            }

            if(joy)
            {
               printString(0, 16, "                    \0");
               printString(0, 16, tempString);
               delay(100);
            }
         }

         waitpadup();

         if(joy & J_A) //do it
         {
            //copy bank 0 to selected bank
            tempString[19] = 0;  //zero out the end of the name
            strcpy(SRAMInfo[RAMposition].ramName, tempString);
            strcpy(SRAMInfo[0].ramName, tempString);
            copyRAMbanks(RAMposition*4, 0, 4);

            saveEverythingToRAM();//backup all info to RAM
         }

         joy = 0;
         printString(0, 14, "                    \0");
         printString(0, 15, "                    \0");
         printString(0, 16, "                    \0");
         printString(0, 17, "                    \0");

         printString(2, 6, "                    \0");
         printString(2, 6, SRAMInfo[0].ramName);
         printString(2, 7, "--------------------\0");
         for(i = 1; i < 4; i++)
         {
            printString(2, 7+i, "                    \0");
            printString(2, 7+i, SRAMInfo[i].ramName);
         }

      }
      else if(joy & J_B)
      {
         waitpadup();
         printString(0, 15, "Press A to load\0");
         printString(0, 16, SRAMInfo[RAMposition].ramName);

         joy = 0;
         while(!(joy & (J_B | J_A | J_START | J_SELECT)))
            joy = joypad();
         waitpadup();

         if(joy & J_A) //do it
         {
            //copy the selected bank to bank 0
            copyRAMbanks(0, RAMposition*4, 4);
            strcpy(SRAMInfo[0].ramName, SRAMInfo[RAMposition].ramName);

	    saveEverythingToRAM();//backup all info to RAM
         }

         joy = 0;
         printString(0, 15, "                    \0");
         printString(0, 16, "                    \0");

         printString(2, 6, "                    \0");
         printString(2, 6, SRAMInfo[0].ramName);
         printString(2, 7, "--------------------\0");
         for(i = 1; i < 4; i++)
         {
            printString(2, 7+i, "                    \0");
            printString(2, 7+i, SRAMInfo[i].ramName);
         }
      }

      if(joy)
         delay(250);
   }
   waitpadup();
   //set_bkg_palette(0, 1, paletteCGB);  //make sure we set it back
}

//we put this in-line in main()
/*char getRomNameStrings()
{
   char numFound, ramOffset;
   char i, j;

   UBYTE *gameName = (UBYTE*)0x4134;
   UBYTE *ninLogo = (UBYTE*)0x4104;

   UBYTE *romSize = (UBYTE*)0x4148;
   //0x07=4076k,0x06=2048k,0x05=1024,0x04=512k,0x02=128k,0x01=64k,0x00=32k

   UBYTE *ramSize = (UBYTE*)0x4149;//0x00=0k, 0x02=8k, 0x03=32k
   UBYTE *realNinLogo = (UBYTE*)0x0104;//this cart's logo (better be right)


   numFound = 0;
   ramOffset = 0;

   i = 1;   //0 should be menu rom.  (okay test case to set it to 1)
   do
   {
      SWITCH_ROM_MBC5((UWORD)i*2);

      for(j = 0; j < 0x30; j++)//check 0x30 bytes
      	if(realNinLogo[j] != ninLogo[j])
           break;

      if(j == 0x30)//it will be <0x30 if they didn't match
      {
         numFound++;

         strncpy(romInfo[numFound].romName, gameName, 16);
         romInfo[numFound].ROMoffset = i;

         if(*ramSize != 0)
         {
            romInfo[numFound].cartRAM = 0x80;	//Cart RAM($80)/No Cart RAM($00)

            romInfo[numFound].RAMoffset = ramOffset;
               //RAM bank 0 offset in SRAM (8k byte increments)

            if(*ramSize == 0x02)
            {
               ramOffset += 1;
               romInfo[numFound].RAMsize = 0x00;	//RAM Usage (0=8k,3=32k)
            }
            else  //otherwise, assume 32k
            {
               ramOffset += 4;
               romInfo[numFound].RAMsize = 0x03;	//RAM Usage (0=8k,3=32k)
            }
         }
         else
         {
            romInfo[numFound].cartRAM = 0;	//Cart RAM($80)/No Cart RAM($00)
            romInfo[numFound].RAMsize = 0;	//RAM Usage (0=8k,3=32k)

            romInfo[numFound].RAMoffset = ramOffset;	//really shouldn't be used, but zero it
         }

	 //ROM Usage (0=32k,1=65k,3=128k,7=256k,f=512k,1f=1M,3f=2M)
         romInfo[numFound].cartROM = ( 0xFF ^ (0xFF << *romSize) );
      }

      i++;

      //for now, we'll have a max of 15 ROMs
   } while((i < 127) && (numFound < 15));  //32M cart is 8 times larger than the PV (127)

   SWITCH_ROM_MBC5((UWORD)1);

   return numFound;
}*/


void printTitle()
{
   char tempString[30];

   printString(0, 0, "Welcome to Flavor's");
   sprintf(tempString, " EMS Menu v%d.%d", VER_MAJ, VER_MIN);
   printString(0, 1, tempString);
}


void printMenu(char numRoms)
{
   char i;
   char tempString[30];
   char tempString2[30];
   UWORD paletteCGB2[4];

   cls();
   hideSprites();

   printTitle();

   for(i = 1; i <= numRoms; i++)
   {
      if(i < 10)
         sprintf(tempString, "0%d-%s", i, romInfo[i].romName);
      else
         sprintf(tempString, "%d-%s", i, romInfo[i].romName);

      printString(1, i+MENU_STARTY, tempString);
   }

   //the palette is now a menu item

   printString(1, 2, "SRAM");

   printString(10, 2, "Colors");

   set_sprite_palette(0, 1, paletteCGB);   //the dynamic palette
   for(i = 0; i < 4; i++)
   {
      set_sprite_tile(i, i + 128); //tile n=n
      set_sprite_prop(i, 0); //palette 0

      paletteCGB2[i] = paletteCGB[0];

      move_sprite(i, (17+i)*8, 4*8);//draw a little palette after "Colors"
   }
   set_sprite_prop(0,1);//gonna use another palette for this one
   set_sprite_tile(0, 129); //spr0 is gonna be tile 129
   set_sprite_palette(1, 1, paletteCGB2);   //the hack palette for pal sprite 0
}

void setPalettes(int doBkg)
{
   unsigned i;

   for(i = 0; i < 8; i++)
   {
      if(doBkg)
         set_bkg_palette(i, 1, paletteCGB);

      set_sprite_palette(i, 1, paletteCGB);
   }

   SHOW_SPRITES;
}

int paletteChooser()
{
   int joy = 0;
   char tempString[30];
   int r[4]={0};
   int g[4]={0};
   int b[4]={0};
   char i, pal, sub, x, y;
   UWORD paletteCGB2[4];
   UWORD savePalette[4];

   cls();
   hideSprites();

   set_bkg_palette(0, 1, hardPaletteCGB);  //make sure we can read stuff
   set_sprite_palette(0, 1, paletteCGB);   //the dynamic palette

   for(i = 0; i < 4; i++)
   {
      set_sprite_tile(i, i + 128); //tile n=n
      set_sprite_prop(i, 0); //palette 0

      set_sprite_tile(i+4, i + 128); //tile n=n
      set_sprite_prop(i+4, 0); //palette 0

      r[i] = paletteCGB[i] & 0x1F;
      g[i] = (paletteCGB[i] >> 5) & 0x1F;
      b[i] = (paletteCGB[i] >> 10) & 0x1F;

      paletteCGB2[i] = paletteCGB[0];

      move_sprite(i+4, (9+i)*8, 19*8);//draw a little palette at bottom
   }
   set_sprite_prop(0,1);//gonna use another palette for this one
   set_sprite_tile(0, 129); //spr0 is gonna be tile 129
   set_sprite_prop(4,1);//gonna use another palette for this one
   set_sprite_tile(4, 129); //spr0 is gonna be tile 129

   set_sprite_palette(1, 1, paletteCGB2);   //the hack palette for pal sprite 0

   memcpy(savePalette, paletteCGB, 4*sizeof(UWORD));

   //printTitle();

   printString(0, 0, "   Palette Editor   \0");
   printString(0, 2, "Exit Keys\0");
   printString(0, 3, "START-save changes  \0");
   printString(0, 4, "SELECT-load defaults\0");
   printString(0, 5, "B-cancel changes    \0");

   move_sprite(0, 2*8, 9*8);
   sprintf(tempString, "R = %d \0", r[0]);
   printString(3, 8, tempString);
   sprintf(tempString, "G = %d \0", g[0]);
   printString(3, 9, tempString);
   sprintf(tempString, "B = %d \0", b[0]);
   printString(3, 10, tempString);

   move_sprite(1, 2*8, 14*8);
   sprintf(tempString, "R = %d \0", r[1]);
   printString(3, 13, tempString);
   sprintf(tempString, "G = %d \0", g[1]);
   printString(3, 14, tempString);
   sprintf(tempString, "B = %d \0", b[1]);
   printString(3, 15, tempString);

   move_sprite(2, 12*8, 9*8);
   sprintf(tempString, "R = %d \0", r[2]);
   printString(13, 8, tempString);
   sprintf(tempString, "G = %d \0", g[2]);
   printString(13, 9, tempString);
   sprintf(tempString, "B = %d \0", b[2]);
   printString(13, 10, tempString);

   move_sprite(3, 12*8, 14*8);
   sprintf(tempString, "R = %d \0", r[3]);
   printString(13, 13, tempString);
   sprintf(tempString, "G = %d \0", g[3]);
   printString(13, 14, tempString);
   sprintf(tempString, "B = %d \0", b[3]);
   printString(13, 15, tempString);

   //SHOW_SPRITES;

   pal = 0; //0-3
   sub = 0; //0=R 1=G 2=B
   x = 2;
   y = 8;
   printString(x, y, arrowString);  //draw new arrow
   while(!(joy & (J_START | J_B | J_SELECT)))
   {
      joy = joypad();

      if(joy & (J_UP | J_DOWN))
      {
	      if(joy & J_DOWN)
   	   {
      	   if(sub == 2)
         	{
	            pal++;
   	         sub = 0;
      	   }
         	else
	         {
   	         sub++;
      	   }

      	   if(pal > 3)
   	         pal = 0;
	      } //move down

      	if(joy & J_UP)
   	   {
	         if(sub == 0)
         	{
      	      pal--;
   	         sub = 2;
	         }
         	else
      	   {
   	         sub--;
	         }

         	if(pal < 0)
      	      pal = 3;

	      } //move up


         printString(x, y, " \0");  //draw new arrow
         switch(pal)
         {
            case 0:
               x = 2;
               y = 8 + sub;
               break;
            case 1:
               x = 2;
               y = 13 + sub;
               break;
            case 2:
               x = 12;
               y = 8 + sub;
               break;
            case 3:
               x = 12;
               y = 13 + sub;
               break;
         }
         printString(x, y, arrowString);  //draw new arrow
      }

      if(joy & (J_LEFT | J_RIGHT))
      {
         if(joy & J_LEFT)
         	i = -1;
         else //if(joy & J_RIGHT)
            i = 1;

         switch(sub)
         {
         	case 0://R
               r[pal] += i;
               if(r[pal] < 0)
               	r[pal] = 31;  //roll back
               else if(r[pal] > 31)
                  r[pal] = 0;   //roll over
            	break;
         	case 1://G
               g[pal] += i;
               if(g[pal] < 0)
               	g[pal] = 31;  //roll back
               else if(g[pal] > 31)
                  g[pal] = 0;   //roll over
            	break;
         	case 2://B
               b[pal] += i;
               if(b[pal] < 0)
               	b[pal] = 31;  //roll back
               else if(b[pal] > 31)
                  b[pal] = 0;   //roll over
            	break;
         }

         switch(pal)
         {
	         case 0:
			   sprintf(tempString, "R = %d \0", r[0]);
			   printString(3, 8, tempString);
		   	sprintf(tempString, "G = %d \0", g[0]);
			   printString(3, 9, tempString);
			   sprintf(tempString, "B = %d \0", b[0]);
			   printString(3, 10, tempString);
         	break;

            case 1:
			   sprintf(tempString, "R = %d \0", r[1]);
			   printString(3, 13, tempString);
			   sprintf(tempString, "G = %d \0", g[1]);
		   	printString(3, 14, tempString);
			   sprintf(tempString, "B = %d \0", b[1]);
			   printString(3, 15, tempString);
            break;

            case 2:
			   sprintf(tempString, "R = %d \0", r[2]);
			   printString(13, 8, tempString);
			   sprintf(tempString, "G = %d \0", g[2]);
		   	printString(13, 9, tempString);
			   sprintf(tempString, "B = %d \0", b[2]);
			   printString(13, 10, tempString);
            break;

            case 3:
			   sprintf(tempString, "R = %d \0", r[3]);
			   printString(13, 13, tempString);
			   sprintf(tempString, "G = %d \0", g[3]);
		   	printString(13, 14, tempString);
			   sprintf(tempString, "B = %d \0", b[3]);
			   printString(13, 15, tempString);
            break;
         }

         paletteCGB[pal] = RGB(r[pal], g[pal], b[pal]);//set the new palette
         if(pal == 0)
         {
            for(i = 0; i < 4; i++)
            {
               paletteCGB2[i] = paletteCGB[0];
            }
            set_sprite_palette(1, 1, paletteCGB2);  //have to use a diff palette for sprite 0
         }
         else
         {
            set_sprite_palette(0, 1, paletteCGB);   //the dynamic palette
         }

      }//if Left or Right

      if(joy)
      	delay(250);
   }
   waitpadup();

   hideSprites();

   if(joy & J_B)
   {
      memcpy(paletteCGB, savePalette, 4*sizeof(UWORD));//fix it
      setPalettes(0);

      return 0;//not successful
   }
   else if(joy & J_START)  //start saves the paletteCGB (already ready)
   {
      setPalettes(0);

      savePalToRAM();//save the palette
      return 1;//successful
   }
   else //if(joy & J_SELECT)
   {
      memcpy(paletteCGB, hardPaletteCGB, 4*sizeof(UWORD));//load the default
      setPalettes(0);

      savePalToRAM();//save the palette
      return 1;//successful
   }
}

void main(void)
{
   int joy;
   int ROMNumber = 1;
   char numRoms;//this will hold the number of roms that we found
   char i, j;
   char saveBuffer[30];

   char tempString[30];
   char *RAM_DF00;
   //char *RAM_BF00;

   RAM_DF00 = (char *)0xDF00;

   SWITCH_ROM_MBC5((UWORD)1);

   cls();
   hideSprites();

   loadSRAMInfo();//get the Pal and Names from RAM

   VBK_REG = 0;
   set_bkg_data( 0, 132, terminalTiles );

   for(i = 0; i < 8; i++)
   {
      set_bkg_palette(i, 1, hardPaletteCGB);
      set_sprite_palette(i, 1, paletteCGB);
   }

   SPRITES_8x8;
   DISPLAY_ON;
   SHOW_BKG;

   setPalettes(0);


      cls();
      for(i = 0; i < 14; i++)
	      printString(0, i, SRAMInfo[i].ramName);
      waitJoyPress();
      cls();


   //the first thing that we should do is go collect all the info about the ROMs


//numRoms = getRomNameStrings();
{
   char numFound, ramOffset;
   char i, j;

   char  tempRomName[20];
   UBYTE tempCartRAM;	//Cart RAM($80)/No Cart RAM($00)
   UBYTE tempCartROM;	//ROM Usage (0=32k,1=65k,3=128k,7=256k,f=512k,1f=1M,3f=2M)
   UBYTE tempRAMsize;	//RAM Usage (0=8k,3=32k)
   UBYTE tempROMsize;


   UBYTE *gameName = (UBYTE*)0x4134;
   UBYTE *ninLogo = (UBYTE*)0x4104;

   UBYTE *romSize = (UBYTE*)0x4148;
   //0x07=4076k,0x06=2048k,0x05=1024,0x04=512k,0x02=128k,0x01=64k,0x00=32k

   UBYTE *ramSize = (UBYTE*)0x4149;//0x00=0k, 0x02=8k, 0x03=32k
   UBYTE *realNinLogo = (UBYTE*)0x0104;//this cart's logo (better be right)


   numFound = 0;
   ramOffset = 4;  //for the EMS carts, we need to start this at 4, because we need 0-3 for working SRAM
                   //that also means that we only get banks 4-14, so we may run out
   ramOverflow = 0;//assume the best

   i = 1;   //0 should be menu rom.  (okay test case to set it to 1)
   do
   {
      SWITCH_ROM_MBC5((UWORD)i*2);

      for(j = 0; j < 0x30; j++)//check 0x30 bytes
      	if(realNinLogo[j] != ninLogo[j])
           break;

      strncpy(tempRomName, gameName, 16);
      tempRAMsize = *ramSize;
      tempROMsize = *romSize;
      SWITCH_ROM_MBC5((UWORD)1);//switch back

      if(j == 0x30)//it will be <0x30 if they didn't match
      {
         //special test case for 16M carts
         if(i == 64)
         {
            if(strncmp(tempRomName, (UBYTE*)0x0134, 16) == 0)
            {  //if the 64th ROM is the menu rom
               break;  //break out of the while loop
            }
         }

         numFound++;

         strncpy(romInfo[numFound].romName, tempRomName, 16);
         romInfo[numFound].ROMoffset = i;

         romInfo[numFound].RAMoffset = ramOffset;//RAM bank 0 offset in SRAM (8k byte increments)

         if(tempRAMsize != 0)
         {
            romInfo[numFound].cartRAM = 0x80;	//Cart RAM($80)/No Cart RAM($00)

            if(tempRAMsize == 0x02)
            {
               if(ramOffset < 0xF)//can't pass 14
                  ramOffset += 1;
               else
                  ramOverflow = 1;

               romInfo[numFound].RAMsize = 0x00;	//RAM Usage (0=8k,3=32k)
            }
            else  //otherwise, assume 32k
            {
               if(ramOffset < 11)//can't pass 11
                  ramOffset += 4;
               else
                  ramOverflow = 1;
               romInfo[numFound].RAMsize = 0x03;	//RAM Usage (0=8k,3=32k)
            }
         }
         else
         {
            romInfo[numFound].cartRAM = 0;	//Cart RAM($80)/No Cart RAM($00)
            romInfo[numFound].RAMsize = 0;	//RAM Usage (0=8k,3=32k)
         }

	      //ROM Usage (0=32k,1=65k,3=128k,7=256k,f=512k,1f=1M,3f=2M)
         romInfo[numFound].cartROM = ( 0xFF ^ (0xFF << tempROMsize) );
      }

      i++;

      //for now, we'll have a max of 15 ROMs

      //on the 32M cart, i should go from 1-127
      //if you use a 16M cart, i should go from 1-63
      //so, if 64 is the menu ROM, then we should quit.
   } while((i < 127) && (numFound < 15));  //32M cart is 8 times larger than the PV (127)

   numRoms = numFound;
}

   {
      for(i = 1; i < 16; i++)
         if(strncmp(savedRomInfo[i].romName, romInfo[i].romName, 20) != 0)
            break;

      if(i != 16)
      {
         printString(0, 1, "New ROM loaded.");
      }

      sprintf(tempString, "i=%d", i);
      printString(0, 4, tempString);

      printString(0, 6, "savedRomInfo\0");
      printString(0, 7, savedRomInfo[i].romName);

      printString(0, 9, "romInfo\0");
      printString(0,10, romInfo[i].romName);

      waitJoyPress();//wait for a button to be pressed
   }

   printMenu(numRoms);

   if(numRoms == 0)
   {
      printString(0, 8, "No ROMs found!!!");
      ROMNumber = 0;  //otherwise, it'll be 1
//      return;
   }

   if(ROMNumber == 0)
	   printString(9, 2, arrowString);  //draw new arrow
   else if(ROMNumber == -1)
	   printString(0, 2, arrowString);  //draw new arrow
   else
	   printString(0,ROMNumber+MENU_STARTY, arrowString);


   do
   {
      joy = joypad();

      if(joy & (J_SELECT | J_UP | J_DOWN))
      {
         if(ROMNumber == 0)  //the palette chooser
            printString(9, 2, " \0");  //delete arrow
         else if(ROMNumber == -1)
            printString(0, 2, " \0");  //delete arrow
         else
            printString(0,ROMNumber+MENU_STARTY, " \0");  //erase old arrow

         if(joy & (J_SELECT | J_DOWN))
            ROMNumber++;
         else if(joy & J_UP)
            ROMNumber--;

         if(ROMNumber < -1)               //1 is the first, but palette is #0
            ROMNumber = numRoms;         //roll over to the last
         else if(ROMNumber > numRoms)    //numRoms-1 is the last
            ROMNumber = -1;               //roll back to the 1st

         if(ROMNumber == 0)  //palette chooser
            printString(9, 2, arrowString);  //draw new arrow
         else if(ROMNumber == -1)  //SRAM editor
            printString(0, 2, arrowString);  //draw new arrow
         else
            printString(0,ROMNumber+MENU_STARTY, arrowString);

         delay(250);
      }

      if((ROMNumber == 0) && (joy & (J_A | J_START)) )
      {
         waitpadup();

         paletteChooser();

         printMenu(numRoms);
         printString(9, 2, arrowString);  //draw new arrow

         joy = 0;  //eat the keypress
      }

      if((ROMNumber == -1) && (joy & (J_A | J_START)) )
      {
         waitpadup();
         ramMenu();

         printMenu(numRoms);
         printString(0, 2, arrowString);  //draw new arrow

         joy = 0;  //eat the keypress
      }


   } while (!(joy & (J_A | J_START)));


   //waitpadup();

   hideSprites();
   cls();

   printString(0, 3, "Release keys to load\0");
   printString(0, 5, romInfo[ROMNumber].romName);

/*  DEBUG
   sprintf(tempString, "  cartROM=0x%x", romInfo[ROMNumber].cartROM);
   printString(0, 8, tempString);
   sprintf(tempString, "  RAMsize=0x%x", romInfo[ROMNumber].RAMsize);
   printString(0, 9, tempString);
   sprintf(tempString, "  RAMoffset=0x%x", romInfo[ROMNumber].RAMoffset);
   printString(0, 10, tempString);
   sprintf(tempString, "  ROMoffset=0x%x", romInfo[ROMNumber].ROMoffset);
   printString(0, 11, tempString);
   sprintf(tempString, "  cartRAM=0x%x", romInfo[ROMNumber].cartRAM);
   printString(0, 12, tempString);
*/

   if(romInfo[ROMNumber].cartRAM == 0)
   {
      printString(0, 7, "ROM uses no SRAM.\0");
   }
   else if(romInfo[ROMNumber].RAMsize == 0x00)
   {
      printString(0, 7, "ROM uses 8kB SRAM.\0");
   }
   else  //otherwise, assume 32k
   {
      printString(0, 7, "ROM uses 32kB SRAM.\0");
   }

   printString(0, 15,   "Questions/Comments? \0");
   printString(0, 16,   "Contact Flavor at   \0");
   printString(0, 17,   "Flavor@Flavor.8k.com\0");

   //while(!joypad()) {}
   waitpadup();
   cls();




   //if the cart uses SRAM, save it's name as SRAMInfo[0].ramName
   if(romInfo[ROMNumber].cartRAM != 0)//cart uses SRAM
   {
      doSRamSwap(ROMNumber);//only do the swap if this one uses SRAM
   }
   else
   {
      backupSRam();  //back it up  (just in case)
   }


   
      for(i = 0; i < 14; i++)
	      printString(0, i, SRAMInfo[i].ramName);
      waitJoyPress();
      cls();


   DISPLAY_OFF;
   HIDE_BKG;
   HIDE_SPRITES;

   saveRomAndRamInfo();
   saveLastRunROMNumber(ROMNumber);

   setPalettes(TRUE);//set the sprite AND bkg palettes
   set_EMS_Offset(romInfo[ROMNumber].ROMoffset);
}
